module Data.Text.Prettyprint.Doc.Internal.Debug {-# DEPRECATED "Use \"Prettyprinter.Internal.Debug\" instead." #-} (
    module Prettyprinter.Internal.Debug
) where

import Prettyprinter.Internal.Debug
