module Data.Text.Prettyprint.Doc.Symbols.Unicode {-# DEPRECATED "Use \"Prettyprinter.Symbols.Unicode\" instead." #-} (
    module Prettyprinter.Symbols.Unicode
) where

import Prettyprinter.Symbols.Unicode
