module Data.Text.Prettyprint.Doc.Render.Tutorials.TreeRenderingTutorial {-# DEPRECATED "Use \"Prettyprinter.Render.Tutorials.TreeRenderingTutorial\" instead." #-} (
    module Prettyprinter.Render.Tutorials.TreeRenderingTutorial
) where

import Prettyprinter.Render.Tutorials.TreeRenderingTutorial
