module Data.Text.Prettyprint.Doc.Render.Util.StackMachine {-# DEPRECATED "Use \"Prettyprinter.Render.Util.StackMachine\" instead." #-} (
    module Prettyprinter.Render.Util.StackMachine
) where

import Prettyprinter.Render.Util.StackMachine



