module Data.Text.Prettyprint.Doc.Render.Tutorials.StackMachineTutorial {-# DEPRECATED "Use \"Prettyprinter.Render.Tutorials.StackMachineTutorial\" instead." #-} (
    module Prettyprinter.Render.Tutorials.StackMachineTutorial
) where

-- Yeah, this produces a deprecation warning. It's hard to disable it while
-- staying compatible with GHC < 8.0 though. Don't waste your time.
import Prettyprinter.Render.Tutorials.StackMachineTutorial
