module Data.Text.Prettyprint.Doc.Internal.Type {-# DEPRECATED "Use \"Prettyprinter.Internal.Type\" instead." #-} (
    module Prettyprinter.Internal.Type
) where

import Prettyprinter.Internal.Type



